@@ -1,4 +1,5 @@
-""""Blocking and non-blocking (asynchronous) clients for CouchDB using Tornado's httpclient.
+""""Blocking and non-blocking (asynchronous) clients for CouchDB using
+Tornado's httpclient.
 
 This module wraps the CouchDB HTTP REST API and defines a common interface
 for making blocking and non-blocking operations on a CouchDB.
@@ -6,10 +7,10 @@ for making blocking and non-blocking operations on a CouchDB.
 
 
 __all__ = ["BlockingCouch", "AsyncCouch", "CouchException", "NotModified",
-        "BadRequest", "NotFound", "MethodNotAllowed", "Conflict",
-        "PreconditionFailed", "InternalServerError"]
+           "BadRequest", "NotFound", "MethodNotAllowed", "Conflict",
+           "PreconditionFailed", "InternalServerError"]
 
-__version__ = '0.2.1'
+__version__ = '0.2.2'
 
 
 import copy
@@ -28,7 +29,7 @@ class AsyncCouch(object):
 
         import couch
         from tornado import ioloop, gen
-        
+
         @gen.coroutine
         def run_test():
             db = couch.AsyncCouch('mytestdb')
@@ -37,7 +38,7 @@ class AsyncCouch(object):
             doc = yield db.get_doc(r['id'])
             yield db.delete_doc(doc)
             yield db.delete_db()
-        
+
         ioloop.IOLoop.run_sync(run_test)
 
     For any methods of this class: If the database call results in an error,
@@ -46,7 +47,7 @@ class AsyncCouch(object):
     yielding to a Future (using gen.coroutine).
     """
 
-    def __init__(self, db_name='', couch_url='http://127.0.0.1:5984',
+    def __init__(self, db_name='', couch_url='http://127.0.0.1:5984/',
                  io_loop=None, **request_args):
         """Creates an `AsyncCouch`.
 
@@ -67,7 +68,10 @@ class AsyncCouch(object):
         """
         self.request_args = request_args
         self._client = httpclient.AsyncHTTPClient(io_loop)
-        self.couch_url = couch_url
+        if couch_url.endswith('/'):
+            self.couch_url = couch_url
+        else:
+            self.couch_url = couch_url + '/'
         self.db_name = db_name
         self._closed = False
 
@@ -84,25 +88,25 @@ class AsyncCouch(object):
     @gen.coroutine
     def create_db(self, db_name=None):
         """Creates a new database."""
-        r = yield self._http_put('/' + (db_name or self.db_name))
+        r = yield self._http_put(db_name or self.db_name)
         raise gen.Return(r)
 
     @gen.coroutine
     def delete_db(self, db_name=None):
         """Deletes the database."""
-        r = yield self._http_delete('/' + (db_name or self.db_name))
+        r = yield self._http_delete(db_name or self.db_name)
         raise gen.Return(r)
 
     @gen.coroutine
     def list_dbs(self):
         """List names of databases."""
-        r = yield self._http_get('/_all_dbs')
+        r = yield self._http_get('_all_dbs')
         raise gen.Return(r)
 
     @gen.coroutine
     def info_db(self, db_name=None):
         """Get info about the database."""
-        r = yield self._http_get('/' + (db_name or self.db_name))
+        r = yield self._http_get(db_name or self.db_name)
         raise gen.Return(r)
 
     @gen.coroutine
@@ -115,13 +119,13 @@ class AsyncCouch(object):
             'target': (db_name or self.db_name),
             'create_target': create_target
         })
-        r = yield self._http_post('/_replicate', body, request_timeout=120.0)
+        r = yield self._http_post('_replicate', body, request_timeout=120.0)
         raise gen.Return(r)
 
     @gen.coroutine
     def uuids(self, count=1):
         """Get one or more uuids."""
-        r = yield self._http_get('/_uuids?count={0}'.format(count))
+        r = yield self._http_get('_uuids?count={0}'.format(count))
         raise gen.Return(r['uuids'])
 
     #
@@ -131,10 +135,19 @@ class AsyncCouch(object):
     @gen.coroutine
     def get_doc(self, doc_id):
         """Get document with the given `doc_id`."""
-        url = '/{0}/{1}'.format(self.db_name, url_escape(doc_id))
+        url = '{0}/{1}'.format(self.db_name, url_escape(doc_id))
         r = yield self._http_get(url)
         raise gen.Return(r)
 
+    @gen.coroutine
+    def has_doc(self, doc_id):
+        """Check if document with the given `doc_id` exists.
+        Returns True if document exists, returns False otherwise.
+        """
+        url = '{0}/{1}'.format(self.db_name, url_escape(doc_id))
+        r = yield self._http_head(url)
+        raise gen.Return(r['code'] == 200)
+
     @gen.coroutine
     def get_docs(self, doc_ids):
         """Get multiple documents with the given list of `doc_ids`.
@@ -145,7 +158,7 @@ class AsyncCouch(object):
         If one or more documents are not found in the database, a NotFound
         exception is raised.
         """
-        url = '/{0}/_all_docs?include_docs=true'.format(self.db_name)
+        url = '{0}/_all_docs?include_docs=true'.format(self.db_name)
         body = json_encode({'keys': doc_ids})
         r = yield self._http_post(url, body)
         raise gen.Return([row['doc'] for row in r['rows']])
@@ -156,13 +169,13 @@ class AsyncCouch(object):
         with id and rev of the saved doc.
         """
         body = json_encode(doc)
-        if '_id' in doc and '_rev' in doc:
-            # update an existing document
-            url = '/{0}/{1}'.format(self.db_name, url_escape(doc['_id']))
+        if '_id' in doc:
+            # create new document, or update an existing document
+            url = '{0}/{1}'.format(self.db_name, url_escape(doc['_id']))
             r = yield self._http_put(url, body)
         else:
-            # save a new document
-            url = '/' + self.db_name
+            # create a new document
+            url = self.db_name
             r = yield self._http_post(url, body)
         raise gen.Return(r)
 
@@ -172,32 +185,37 @@ class AsyncCouch(object):
         Response is a list of dicts with id and rev of the saved docs.
         """
         # use bulk docs API to update the docs
-        url = '/{0}/_bulk_docs'.format(self.db_name)
+        url = '{0}/_bulk_docs'.format(self.db_name)
         body = json_encode({'all_or_nothing': all_or_nothing, 'docs': docs})
         r = yield self._http_post(url, body)
         raise gen.Return(r)
 
     @gen.coroutine
     def delete_doc(self, doc):
-        """Delete a document."""
+        """Delete a document.
+        The `doc` shall be a dict, at least having the keys `_id` and `_rev`.
+        """
         if '_rev' not in doc or '_id' not in doc:
             raise KeyError('Missing id or revision information in doc')
-        url = '/{0}/{1}?rev={2}'.format(
+        url = '{0}/{1}?rev={2}'.format(
             self.db_name, url_escape(doc['_id']), doc['_rev'])
         r = yield self._http_delete(url)
         raise gen.Return(r)
 
     @gen.coroutine
     def delete_docs(self, docs, all_or_nothing=False):
-        """Delete multiple documents."""
+        """Delete multiple documents.
+        The `docs` shall be an array of dicts, each at least having the keys
+        `_id` and `_rev`.
+        """
         if any('_rev' not in doc or '_id' not in doc for doc in docs):
             raise KeyError('Missing id or revision information in one or '
                            'more docs')
         # make list of docs to mark as deleted
         deleted = [{'_id': doc['_id'], '_rev': doc['_rev'],
-                '_deleted': True} for doc in docs]
+                    '_deleted': True} for doc in docs]
         # use bulk docs API to update the docs
-        url = '/{0}/_bulk_docs'.format(self.db_name)
+        url = '{0}/_bulk_docs'.format(self.db_name)
         body = json_encode({'all_or_nothing': all_or_nothing,
                             'docs': deleted})
         r = yield self._http_post(url, body)
@@ -221,8 +239,8 @@ class AsyncCouch(object):
                                  ' the given name')
             else:
                 mimetype = doc['_attachments'][attachment_name]['content_type']
-        url = '/{0}/{1}/{2}'.format(self.db_name, url_escape(doc['_id']),
-                                    url_escape(attachment_name))
+        url = '{0}/{1}/{2}'.format(self.db_name, url_escape(doc['_id']),
+                                   url_escape(attachment_name))
         headers = {'Accept': mimetype}
         r = yield self._http_get(url, headers=headers)
         raise gen.Return(r)
@@ -236,7 +254,7 @@ class AsyncCouch(object):
         if any(key not in attachment for key in ['mimetype', 'name', 'data']):
             raise KeyError('Attachment dict is missing one or more '
                            'required keys')
-        url = '/{0}/{1}/{2}{3}'.format(
+        url = '{0}/{1}/{2}{3}'.format(
             self.db_name, url_escape(doc['_id']),
             url_escape(attachment['name']),
             '?rev={0}'.format(doc['_rev']) if '_rev' in doc else '')
@@ -252,9 +270,9 @@ class AsyncCouch(object):
         if '_rev' not in doc or '_id' not in doc:
             raise KeyError('Missing id or revision information in doc')
         else:
-            url = '/{0}/{1}/{2}?rev={3}'.format(self.db_name,
-                    url_escape(doc['_id']), url_escape(attachment_name),
-                    doc['_rev'])
+            url = '{0}/{1}/{2}?rev={3}'.format(
+                self.db_name, url_escape(doc['_id']),
+                url_escape(attachment_name), doc['_rev'])
         r = yield self._http_delete(url)
         raise gen.Return(r)
 
@@ -283,12 +301,12 @@ class AsyncCouch(object):
           limit=<number of docs>
 
         Prevent CouchDB from refreshing a stale view:
-          stale='ok'
-          stale='update_after'
+          stale="ok"
+          stale="update_after"
 
         Reverse the output:
-          descending=true
-          descending=false  (default value)
+          descending=True
+          descending=False  (default value)
 
         Note that the descending option is applied before any key filtering, so
         you may need to swap the values of the startkey and endkey options to
@@ -299,29 +317,29 @@ class AsyncCouch(object):
 
         The group option controls whether the reduce function reduces to a set
         of distinct keys or to a single result row:
-          group=true
-          group=false  (default value)
+          group=True
+          group=False  (default value)
 
           group_level=<number>
 
         Use the reduce function of the view:
-          reduce=true  (default value)
-          reduce=false
+          reduce=True  (default value)
+          reduce=False
 
-        Note that default value of reduce is true, only if a reduce function is
+        Note that default value of reduce is True, only if a reduce function is
         defined for the view.
 
         Automatically fetch and include the document which emitted each view
         entry:
-          include_docs=true
-          include_docs=false  (default value)
+          include_docs=True
+          include_docs=False  (default value)
 
         Determine whether the endkey is included in the result:
-          inclusive_end=true  (default value)
-          inclusive_end=false
+          inclusive_end=True  (default value)
+          inclusive_end=False
         """
-        url = '/{0}/_design/{1}/_view/{2}'.format(self.db_name,
-                design_doc_name, view_name)
+        url = '{0}/_design/{1}/_view/{2}'.format(
+            self.db_name, design_doc_name, view_name)
         r = yield self._view(url, **kwargs)
         raise gen.Return(r)
 
@@ -330,7 +348,7 @@ class AsyncCouch(object):
         """Query the _all_docs view.
         Accepts the same keyword parameters as `view()`.
         """
-        url = '/{0}/_all_docs'.format(self.db_name)
+        url = '{0}/_all_docs'.format(self.db_name)
         r = yield self._view(url, **kwargs)
         raise gen.Return(r)
 
@@ -339,7 +357,7 @@ class AsyncCouch(object):
         """Query a temporary view.
         The view_doc parameter is a dict with the view's map and reduce
         functions."""
-        url = '/{0}/_temp_view'.format(self.db_name)
+        url = '{0}/_temp_view'.format(self.db_name)
         r = yield self._view(url, body=view_doc, **kwargs)
         raise gen.Return(r)
 
@@ -378,12 +396,12 @@ class AsyncCouch(object):
             for item in obj:
                 if 'error' in item:
                     raise relax_exception(httpclient.HTTPError(
-                            resp.code if item['error'] != 'not_found' else 404,
-                            item['reason'], resp))
+                        resp.code if item['error'] != 'not_found' else 404,
+                        item['reason'], resp))
 
         elif 'error' in obj:
-            raise relax_exception(httpclient.HTTPError(resp.code,
-                    obj['reason'], resp))
+            raise relax_exception(httpclient.HTTPError(
+                resp.code, obj['reason'], resp))
 
         elif 'rows' in obj:
             # check if there is an error in the result rows,
@@ -391,14 +409,22 @@ class AsyncCouch(object):
             for row in obj['rows']:
                 if 'error' in row:
                     raise relax_exception(httpclient.HTTPError(
-                            resp.code if row['error'] != 'not_found' else 404,
-                            row['error'], resp))
+                        resp.code if row['error'] != 'not_found' else 404,
+                        row['error'], resp))
         return obj
 
+    def _parse_headers(self, resp):
+        headers = {"code": resp.code}
+        headers.update(resp.headers)
+        return headers
+
+    def _test_closed(self):
+        if self._closed:
+            raise CouchException('Database connection is closed.')
+
     @gen.coroutine
     def _http_get(self, uri, headers=None):
-        if self._closed:
-            raise CouchException('Database connection is closed.')
+        self._test_closed()
         if headers is None:
             headers = {}
         req_args = copy.deepcopy(self.request_args)
@@ -418,16 +444,15 @@ class AsyncCouch(object):
                 raise relax_exception(e)
             resp = e.response
         raise gen.Return(self._parse_response(resp) if decode else resp.body)
-        
+
     @gen.coroutine
     def _http_post(self, uri, body, **kwargs):
-        if self._closed:
-            raise CouchException('Database connection is closed.')
+        self._test_closed()
         req_args = copy.deepcopy(self.request_args)
         req_args.update(kwargs)
         req_args.setdefault('headers', {}).update({
-                'Accept': 'application/json',
-                'Content-Type': 'application/json'})
+            'Accept': 'application/json',
+            'Content-Type': 'application/json'})
         req = httpclient.HTTPRequest(self.couch_url + uri, method='POST',
                                      body=body, **req_args)
         try:
@@ -440,8 +465,7 @@ class AsyncCouch(object):
 
     @gen.coroutine
     def _http_put(self, uri, body='', headers=None):
-        if self._closed:
-            raise CouchException('Database connection is closed.')
+        self._test_closed()
         if headers is None:
             headers = {}
         req_args = copy.deepcopy(self.request_args)
@@ -462,11 +486,10 @@ class AsyncCouch(object):
 
     @gen.coroutine
     def _http_delete(self, uri):
-        if self._closed:
-            raise CouchException('Database connection is closed.')
+        self._test_closed()
         req_args = copy.deepcopy(self.request_args)
         req_args.setdefault('headers', {}).update({
-                'Accept': 'application/json'})
+            'Accept': 'application/json'})
         req = httpclient.HTTPRequest(self.couch_url + uri, method='DELETE',
                                      **req_args)
         try:
@@ -477,6 +500,21 @@ class AsyncCouch(object):
             resp = e.response
         raise gen.Return(self._parse_response(resp))
 
+    @gen.coroutine
+    def _http_head(self, uri):
+        if self._closed:
+            raise CouchException('Database connection is closed.')
+        req_args = copy.deepcopy(self.request_args)
+        req = httpclient.HTTPRequest(self.couch_url + uri, method='HEAD',
+            **req_args)
+        try:
+            resp = yield self._client.fetch(req)
+        except httpclient.HTTPError as e:
+            if not e.response:
+                raise relax_exception(e)
+            resp = e.response
+        raise gen.Return(self._parse_headers(resp))
+
 
 class BlockingCouch(AsyncCouch):
     """Basic wrapper class for blocking operations on a CouchDB.
@@ -498,7 +536,7 @@ class BlockingCouch(AsyncCouch):
     seperate IOLoop.
     """
 
-    def __init__(self, db_name='', couch_url='http://127.0.0.1:5984',
+    def __init__(self, db_name='', couch_url='http://127.0.0.1:5984/',
                  **request_args):
         """Creates a `BlockingCouch`.
 
@@ -550,60 +588,73 @@ class BlockingCouch(AsyncCouch):
 
 class CouchException(httpclient.HTTPError):
     """Base class for Couch specific exceptions"""
+
     def __init__(self, HTTPError, msg=None):
-        httpclient.HTTPError.__init__(self, HTTPError.code,
-                msg, HTTPError.response)
+        httpclient.HTTPError.__init__(
+            self, HTTPError.code, msg, HTTPError.response)
 
 
 class NotModified(CouchException):
     """HTTP Error 304 (Not Modified)"""
+
     def __init__(self, HTTPError):
-        CouchException.__init__(self, HTTPError,
-                'The document has not been modified since the last update.')
+        CouchException.__init__(
+            self, HTTPError,
+            'The document has not been modified since the last update.')
 
 
 class BadRequest(CouchException):
     """HTTP Error 400 (Bad Request)"""
+
     def __init__(self, HTTPError):
-        CouchException.__init__(self, HTTPError, 'The syntax of the request '
-                'was invalid or could not be processed.')
+        CouchException.__init__(
+            self, HTTPError,
+            'The syntax of the request was invalid or could not be processed.')
 
 
 class NotFound(CouchException):
     """HTTP Error 404 (Not Found)"""
+
     def __init__(self, HTTPError):
-        CouchException.__init__(self, HTTPError,
-                'The requested resource was not found.')
+        CouchException.__init__(
+            self, HTTPError, 'The requested resource was not found.')
 
 
 class MethodNotAllowed(CouchException):
     """HTTP Error 405 (Method Not Allowed)"""
+
     def __init__(self, HTTPError):
-        CouchException.__init__(self, HTTPError, 'The request was made using '
-                'an incorrect request method; for example, a GET was used '
-                'where a POST was required.')
+        CouchException.__init__(
+            self, HTTPError, 'The request was made using an incorrect request '
+            'method; for example, a GET was used where a POST was required.')
 
 
 class Conflict(CouchException):
     """HTTP Error 409 (Conflict)"""
+
     def __init__(self, HTTPError):
-        CouchException.__init__(self, HTTPError, 'The request failed because '
-                'of a database conflict.')
+        CouchException.__init__(
+            self, HTTPError,
+            'The request failed because of a database conflict.')
 
 
 class PreconditionFailed(CouchException):
     """HTTP Error 412 (Precondition Failed)"""
+
     def __init__(self, HTTPError):
-        CouchException.__init__(self, HTTPError, 'Could not create database - '
-                'a database with that name already exists.')
+        CouchException.__init__(
+            self, HTTPError, 'Could not create database - '
+            'a database with that name already exists.')
 
 
 class InternalServerError(CouchException):
     """HTTP Error 500 (Internal Server Error)"""
+
     def __init__(self, HTTPError):
-        CouchException.__init__(self, HTTPError, 'The request was invalid and '
-                'failed, or an error occurred within the CouchDB server that '
-                'prevented it from processing the request.')
+        CouchException.__init__(
+            self, HTTPError, 'The request was invalid and failed, '
+            'or an error occurred within the CouchDB server that '
+            'prevented it from processing the request.')
 
 
 def relax_exception(e):
